# @Author: Ilija Medan
# @Date: May 18, 2021
# @Filename: obsmode.py
# @License: BSD 3-Clause
# @Copyright: Ilija Medan

import numpy as np
import warnings
import fitsio
import collections
from scipy.spatial import cKDTree

from mugatu.exceptions import MugatuError, MugatuWarning
from coordio.utils import radec2wokxy, wokxy2radec

try:
    from sdssdb.peewee.sdss5db import database
    database.set_profile('operations')
    _database = True
except:
    _database = False

from sdssdb.peewee.sdss5db.targetdb import Carton, Category, Magnitude, CartonToTarget, Target
from sdssdb.peewee.sdss5db.targetdb import DesignMode as DesignModeDB
from sdssdb.peewee.sdss5db import catalogdb
from sdssdb.peewee.sdss5db import targetdb


def ang_sep(ra1, dec1, ra2, dec2):
    """
    Returns angular separation between objects.

    Parameters
    ----------
    ra1: float or np.array
        Right ascension(s) of first object(s).

    dec1: float or np.array
        Declination(s) of first object(s).

    ra2: float or np.array
        Right ascension(s) of second object(s).

    dec2: float or np.array
        Declination(s) of second object(s).

    Returns
    -------
    sep: float or np.array
        Angular seperation between objects in degrees.
    """
    ra1 = np.radians(ra1)
    dec1 = np.radians(dec1)
    ra2 = np.radians(ra2)
    dec2 = np.radians(dec2)
    sep = (180 / np.pi) * np.arccos(np.sin(dec1) * np.sin(dec2) +
                                    np.cos(dec1) * np.cos(dec2) * np.cos(ra1 - ra2))
    return sep


def allDesignModes(filename=None, ext=1):
    """Function to return a dictionary with all design modes

    Parameters
    ----------
    
    filename : str
        FITS file name to read array from (default None)

    ext : int or str
        FITS file extension to read array from (default 1)

    Comments
    --------

    Returns an OrderedDict
    
    If filename is not provided, it reads from that file name
    and the given extension. If not, it reads from the database.

    If neither the filename is given nor the db is available, it
    returns None.
"""
    if(filename is not None):
        dmas = fitsio.read(filename, ext=ext)
        dmd = collections.OrderedDict()
        for dma in dmas:
            dmd[dma['label']] = DesignMode()
            dmd[dma['label']].fromarray(dma)
        return(dmd)
            
    if(_database | (filename is not None)):
        desmodes = DesignModeDB.select()
        dmd = collections.OrderedDict()
        for desmode in desmodes:
            dmd[desmode.label] = DesignMode(label=desmode.label)
        return(dmd)

    print("No file name input or db access.")
    return(None)


class DesignMode(object):
    """Class to store parameters for a design mode

    Parameters
    ----------

    label : str
        label for design mode

    Attributes
    ----------

    n_stds_min: dict
        Dictonary with the minimum number of standards for a design
        for each instrument ('APOGEE' and 'BOSS').

    min_stds_fovmetric: dict
        Dictonary wih the FOV metric for the standards in a design
        for each instrument ('APOGEE' and 'BOSS').
        The FOV metric is described by three parameters, the
        nth neighbor to get distances to, the percentle distance
        to calculate and the distace to compare against for validation
        (in mm).

    stds_mags: dict
        Dictonary for the min/max magnitude for the standards in a
        design for each instrument ('APOGEE' and 'BOSS'). Indexes
        correspond to magntidues: [g, r, i, bp, gaia_g, rp, h].

    bright_limit_targets: dict
        Dictonary for the min/max magnitude for the science targets
        in adesign for each instrument ('APOGEE' and 'BOSS'). Indexes
        correspond to magntidues: [g, r, i, bp, gaia_g, rp, h].

    sky_neighbors_targets: dict
        Dictonary for the parameters used to check distance between
        skies and all possible sources in field for each instrument
        ('APOGEE' and 'BOSS'). Distances to targets (r, in arcseconds)
        must be r > R_0 * (lim - mags) ** beta, where mags is G band
        for BOSS and H band for APOGEE. Indexes correspond to:
        [R_0, beta, lim]

    trace_diff_targets: dict
        Dictonary for the maximum magnitude difference allowed between
        fibers next to each ther on the chip for each instrument
        ('APOGEE' and 'BOSS'). Here the magntidue difference is checked
        in the G band for BOSS and H band for APOGEE.
"""
    def __init__(self, label=None):
        if(label is not None):
            self.fromdb(label=label)
        return

    def fromdb(self, label=None):
        """Read in parameters for design mode from db

        Parameters
        ----------
        
        label : str
            name of design mode
"""
        if(_database is False):
            print("No database, cannot read.")
            return

        self.desmode_label = label

        desmode = DesignModeDB.select().where(DesignModeDB.label == self.desmode_label)[0]

        self.n_skies_min = {}
        self.n_skies_min['BOSS'] = desmode.boss_skies_min
        self.n_skies_min['APOGEE'] = desmode.apogee_skies_min
        
        self.min_skies_fovmetric = {}
        self.min_skies_fovmetric['BOSS'] = desmode.boss_skies_fov
        self.min_skies_fovmetric['APOGEE'] = desmode.apogee_skies_fov
        
        self.n_stds_min = {}
        self.n_stds_min['BOSS'] = desmode.boss_stds_min
        self.n_stds_min['APOGEE'] = desmode.apogee_stds_min
        
        self.min_stds_fovmetric = {}
        self.min_stds_fovmetric['BOSS'] = desmode.boss_stds_fov
        self.min_stds_fovmetric['APOGEE'] = desmode.apogee_stds_fov
        
        self.stds_mags = {}
        self.stds_mags['BOSS'] = np.zeros((len(desmode.boss_stds_mags_min), 2), dtype=np.float64)
        self.stds_mags['BOSS'][:, 0] = desmode.boss_stds_mags_min
        self.stds_mags['BOSS'][:, 1] = desmode.boss_stds_mags_max
        self.stds_mags['APOGEE'] = np.zeros((len(desmode.apogee_stds_mags_min), 2), dtype=np.float64)
        self.stds_mags['APOGEE'][:, 0] = desmode.apogee_stds_mags_min
        self.stds_mags['APOGEE'][:, 1] = desmode.apogee_stds_mags_max
        
        self.bright_limit_targets = {}
        self.bright_limit_targets['BOSS'] = np.zeros((len(desmode.boss_bright_limit_targets_min), 2), dtype=np.float64)
        self.bright_limit_targets['BOSS'][:, 0] = desmode.boss_bright_limit_targets_min
        self.bright_limit_targets['BOSS'][:, 1] = desmode.boss_bright_limit_targets_max
        self.bright_limit_targets['APOGEE'] = np.zeros((len(desmode.apogee_bright_limit_targets_min), 2), dtype=np.float64)
        self.bright_limit_targets['APOGEE'][:, 0] = desmode.apogee_bright_limit_targets_min
        self.bright_limit_targets['APOGEE'][:, 1] = desmode.apogee_bright_limit_targets_max
        
        self.sky_neighbors_targets = {}
        self.sky_neighbors_targets['BOSS'] = desmode.boss_sky_neighbors_targets
        self.sky_neighbors_targets['APOGEE'] = desmode.apogee_sky_neighbors_targets
        
        self.trace_diff_targets = {}
        self.trace_diff_targets['APOGEE'] = desmode.apogee_trace_diff_targets
        return

    def toarray(self):
        """Returns an ndarray form of the object
"""
        attrnames = ['n_skies_min', 'min_skies_fovmetric', 'n_stds_min',
                     'min_stds_fovmetric', 'stds_mags',
                     'bright_limit_targets', 'sky_neighbors_targets',
                     'trace_diff_targets']

        dtype = [('label', 'U30')]
        for attrname in attrnames:
            attr = getattr(self, attrname)
            for instrument in attr:
                name = instrument.lower() + '_' + attrname
                if(type(attr[instrument]) == np.ndarray):
                    dtype.append((name, np.float64,
                                  attr[instrument].shape))
                elif(type(attr[instrument]) == list):
                    dtype.append((name, np.float64,
                                  len(attr[instrument])))
                else:
                    dtype.append((name, np.float64))
        
        arr = np.zeros(1, dtype=dtype)
        arr['label'] = self.desmode_label
        for attrname in attrnames:
            attr = getattr(self, attrname)
            for instrument in attr:
                name = instrument.lower() + '_' + attrname
                arr[name] = attr[instrument]

        return(arr[0])

    def fromarray(self, arr=None):
        """Reads an ndarray form of the object created by toarray()
"""
        attrnames = ['n_skies_min', 'min_skies_fovmetric', 'n_stds_min',
                     'min_stds_fovmetric', 'stds_mags',
                     'bright_limit_targets', 'sky_neighbors_targets',
                     'trace_diff_targets']

        self.desmode_label = arr['label']
        for attrname in attrnames:
            setattr(self, attrname, dict())
            for instrument in ['BOSS', 'APOGEE']:
                name = instrument.lower() + '_' + attrname
                if(name in arr.dtype.names):
                    getattr(self, attrname)[instrument] = arr[name]
        
        return

    def todict(self):
        """Returns a dictionary form of the object
"""
        attrs = ['n_skies_min', 'min_skies_fovmetric', 'n_stds_min',
                 'min_stds_fovmetric', 'stds_mags', 'bright_limit_targets',
                 'sky_neighbors_targets', 'trace_diff_targets']

        dmd = dict()
        dmd['label'] = self.desmode_label
        for attr in attrs:
            dmda = getattr(self, attr)
            dmd[attr] = dict()
            for key in dmda:
                if(type(dmda[key]) == np.ndarray):
                    dmd[attr][key] = dmda[key].tolist()
                else:
                    dmd[attr][key] = dmda[key]

        return(dmd)

    def fromdict(self, designmode_dict=None):
        """Builds from a dictionary form of the object

        Parameters
        ----------
        
        designmode_dict : dict
            dictionary form of DesignMode object created by todict()
"""
        attrs = ['n_skies_min', 'min_skies_fovmetric', 'n_stds_min',
                 'min_stds_fovmetric', 'stds_mags', 'bright_limit_targets',
                 'sky_neighbors_targets', 'trace_diff_targets']

        self.desmod_label = designmode_dict['label']
        for param in attrs:
            setattr(self, param, dict())
            for instrument in designmode_dict[param]:
                if(type(designmode_dict[param][instrument]) == list):
                    getattr(self, param)[instrument] = np.array(designmode_dict[param][instrument])
                else:
                    getattr(self, param)[instrument] = designmode_dict[param][instrument]
        return

    def frommanual(self, label=None, desmode_manual=None):
        """Read in parameters for design mode from dictionary input

        Parameters
        ----------
        
        label : str
            name of design mode

        desmode_manual : dict
            Dictonary of DesignMode parameters to be used as manual
            inputs to validate the design, rather than from targetdb.
"""
        self.desmode_label = label

        self.n_skies_min = {}
        self.n_skies_min['BOSS'] = desmode_manual['boss_skies_min']
        self.n_skies_min['APOGEE'] = desmode_manual['apogee_skies_min']
        
        self.min_skies_fovmetric = {}
        self.min_skies_fovmetric['BOSS'] = desmode_manual['boss_skies_fov']
        self.min_skies_fovmetric['APOGEE'] = desmode_manual['apogee_skies_fov']
        
        self.n_stds_min = {}
        self.n_stds_min['BOSS'] = desmode_manual['boss_stds_min']
        self.n_stds_min['APOGEE'] = desmode_manual['apogee_stds_min']
        
        self.min_stds_fovmetric = {}
        self.min_stds_fovmetric['BOSS'] = desmode_manual['boss_stds_fov']
        self.min_stds_fovmetric['APOGEE'] = desmode_manual['apogee_stds_fov']
    
        self.stds_mags = {}
        self.stds_mags['BOSS'] = desmode_manual['boss_stds_mags']
        self.stds_mags['APOGEE'] = desmode_manual['apogee_stds_mags']
        
        self.bright_limit_targets = {}
        self.bright_limit_targets['BOSS'] = desmode_manual['boss_bright_limit_targets']
        self.bright_limit_targets['APOGEE'] = desmode_manual['apogee_bright_limit_targets']

        self.sky_neighbors_targets = {}
        self.sky_neighbors_targets['BOSS'] = desmode_manual['boss_sky_neighbors_targets']
        self.sky_neighbors_targets['APOGEE'] = desmode_manual['apogee_sky_neighbors_targets']

        self.trace_diff_targets = {}
        self.trace_diff_targets['APOGEE'] = desmode_manual['apogee_trace_diff_targets']
        return


def check_assign_mag_limit(mag_metric_min,
                           mag_metric_max,
                           assign_mag):
    """
    Checks the if magnitude of one assignment agrees with
    design mode for some instrument and carton class

    Parameters
    ----------
    mag_metric_min: float
        The minimum magitude for the specific DesignMode.

    mag_metric_min: float
        The maximum magitude for the specific DesignMode.

    assign_mag: float
        The magntiude of the assignment being checkaged against
        the DesignMode.

    Returns
    -------
    targ_check: boolean
        True of the assignment magntiude passed the DesignMode
        for the magnitude, False if not.

    complete_check: str
        'COMPLETE' if assign_mag is value, 'INCOMPLETE' if Null.
    """

    # set default values
    complete_check = 'COMPLETE'
    targ_check = False

    # do checks
    # define Null cases for targetdb.magnitude table
    cases = [-999, -9999, 999,
             0.0, np.nan, 99.9, None]
    if assign_mag in cases or np.isnan(assign_mag):
        complete_check = 'INCOMPLETE'
        # set True, no mag is not a fail
        targ_check = True
    # check when greater than and less than
    elif mag_metric_min != -999. and mag_metric_max != -999.:
        if (mag_metric_min < assign_mag < mag_metric_max):
            targ_check = True
    # check when just greater than
    elif mag_metric_min != -999.:
        if assign_mag > mag_metric_min:
            targ_check = True
    # check when less than
    else:
        if assign_mag < mag_metric_max:
            targ_check = True
    return targ_check, complete_check


def bright_neigh_exclusion_r(mag_bs, mag_limit_r, lunation):
    """
    returns the exclusion radius for a fiber around a bright star
    in arcseconds for a given designmode. This is for the piecewise
    appromixation used based on:
    https://wiki.sdss.org/pages/viewpage.action?pageId=100173069

    Parameters
    ----------
    mag_bs: float or np.array
        The magniutde in the G band for the bright star(s)

    mag_limit_r: float
        Magnitude limit for the designmode in the r-SDSS band

    lunation: str:
        If the designmode is bright time ('bright') or dark
        time ('dark')

    Returns
    -------
    r_exclude: float or np.array
        exclusion radius in arcseconds around bright star(s)
    """
    # linear portion in the wings
    r_wings = (mag_limit_r - mag_bs - 8.2) / 0.05
    # linear portion in transition area
    r_trans = (mag_limit_r - mag_bs - 4.5) / 0.25
    # core area
    if lunation == 'bright':
        r_core = 1.75 * (mag_limit_r - mag_bs) ** 0.6
    else:
        r_core = 1.5 * (mag_limit_r - mag_bs) ** 0.8
    # exlusion radius is the max of each section
    if isinstance(mag_bs, float):
        if mag_bs <= mag_limit_r:
            r_exclude = max(r_wings, r_trans, r_core)
        else:
            r_exclude = 0.
    else:
        r_exclude = np.nanmax(np.column_stack((r_wings,
                                               r_trans,
                                               r_core)),
                              axis=1)
        r_exclude[mag_bs > mag_limit_r] = 0.
    return r_exclude


def adjusted_brigh_neigh_mag(mag_bs, r, lunation):
    """
    returns the approximate adjusted magntidue of a
    bright source at the position of a nearby fiber.
    This is for the piecewise
    appromixation used based on:
    https://wiki.sdss.org/pages/viewpage.action?pageId=100173069

    Parameters
    ----------
    mag_bs: float
        The magniutde in the G band for the bright star

    r: float or np.array
        distance between fiber and bright star (in arcseconds)

    lunation: str:
        If the designmode is bright time ('bright') or dark
        time ('dark')

    Returns
    -------
    adjusted_mag_bs: float
        The adjusted magntiude of the bright star in SDSS r-band
    """
    mag_wings = 8.2 + 0.05 * r + mag_bs
    mag_trans = 4.5 + 0.25 * r + mag_bs
    if lunation == 'bright':
        mag_core = (r / 1.75) ** (1 / 0.6) + mag_bs
    else:
        mag_core = (r / 1.5) ** (1 / 0.8) + mag_bs
    if isinstance(r, float):
        adjusted_mag_bs = np.nanmin([mag_wings,
                                     mag_trans,
                                     mag_core])
    else:
        adjusted_mag_bs = np.nanmin(np.column_stack((mag_wings,
                                                     mag_trans,
                                                     mag_core)),
                                    axis=1)
    return adjusted_mag_bs


def build_brigh_neigh_query(check_type, instrument, mag_lim,
                            racen, deccen):
    """
    Builds the database query needed to run bright
    neighbor check

    Parameters
    ----------
    check_type: str
        Either 'designmode' for full designmode check with
        all stars down to mag_lim in catalogdb, or 'safety'
        for safety check using bright stars in targetdb.

    instrument: str
        Fiber instrument being checked. Either 'BOSS' or
        'APOGEE'.

    mag_lim: float
        Magnitude limit in the r_SDSS band for BOSS or
        H band for APOGEE.

    racen: float
        Field center in right ascension.

    deccen: float
        Feild center in declination

    Outputs
    -------
    db_query_results: tuple
        Tuple of (ra, dec, mag, catalogid) for the
        appropriate database query
    """
    # return empty tuple if no mag limit
    if mag_lim == -999.:
        return ()
    if check_type == 'designmode':
        if instrument == 'BOSS':
            cat = catalogdb.Gaia_DR2
            ra_col = catalogdb.Gaia_DR2.ra
            dec_col = catalogdb.Gaia_DR2.dec
            ra_col_str = 'ra'
            dec_col_str = 'dec'
            mag_col = catalogdb.Gaia_DR2.phot_g_mean_mag
            # run the query
            db_query_gaia = (catalogdb.CatalogToTIC_v8.select(
                ra_col,
                dec_col,
                mag_col,
                catalogdb.CatalogToTIC_v8.catalogid)
                .join(catalogdb.TIC_v8)
                .join(catalogdb.Gaia_DR2)
                .where((cat.cone_search(racen,
                                        deccen,
                                        1.5,
                                        ra_col=ra_col_str,
                                        dec_col=dec_col_str)) &
                       (mag_col < mag_lim)))
            rasg, decsg, magsg, catalogidsg = map(list, zip(*list(db_query_gaia.tuples())))
            rasg = np.array(rasg)
            decsg = np.array(decsg)
            magsg = np.array(magsg)

            cat = catalogdb.Tycho2
            ra_col = catalogdb.Tycho2.radeg
            dec_col = catalogdb.Tycho2.dedeg
            ra_col_str = 'radeg'
            dec_col_str = 'dedeg'
            mag_colbt = catalogdb.Tycho2.btmag
            mag_colvt = catalogdb.Tycho2.vtmag
            # run the query
            db_query_tych = (catalogdb.CatalogToTycho2.select(
                ra_col,
                dec_col,
                mag_colbt,
                mag_colvt,
                catalogdb.CatalogToTycho2.catalogid)
                .join(catalogdb.Tycho2)
                .where((cat.cone_search(racen,
                                        deccen,
                                        1.5,
                                        ra_col=ra_col_str,
                                        dec_col=dec_col_str)) &
                       (mag_colvt < mag_lim)))
            rast, decst, magsbt, magsvt, catalogidst = map(list, zip(*list(db_query_tych.tuples())))
            rast = np.array(rast)
            decst = np.array(decst)
            magsbt = np.array(magsbt)
            magsvt = np.array(magsvt)
            magsg_tych = np.zeros(len(magsbt))
            magsg_tych[magsbt != None] = (magsvt[magsbt != None] - 0.02051 -
                                         0.2706 * (magsbt[magsbt != None] -
                                                   magsvt[magsbt != None]) +
                                         0.03394 * (magsbt[magsbt != None] -
                                                    magsvt[magsbt != None]) ** 2 -
                                         0.05937 * (magsbt[magsbt != None] -
                                                    magsvt[magsbt != None]) ** 3)
            magsg_tych[magsbt == None] = magsvt[magsbt == None] - 1
            ras = np.append(rasg, rast)
            decs = np.append(decsg, decst)
            mags = np.append(magsg, magsg_tych)
            catalogids = catalogidsg + catalogidst
            db_query_results = (ras, decs, mags, catalogids)

        else:
            cat = catalogdb.TwoMassPSC
            ra_col = catalogdb.TwoMassPSC.ra
            dec_col = catalogdb.TwoMassPSC.decl
            ra_col_str = 'ra'
            dec_col_str = 'decl'
            mag_col = catalogdb.TwoMassPSC.h_m
            # run the query
            db_query = (catalogdb.CatalogToTIC_v8.select(
                ra_col,
                dec_col,
                mag_col,
                catalogdb.CatalogToTIC_v8.catalogid)
                .join(catalogdb.TIC_v8)
                .join(cat)
                .where((cat.cone_search(racen,
                                        deccen,
                                        1.5,
                                        ra_col=ra_col_str,
                                        dec_col=dec_col_str)) &
                       (mag_col < mag_lim)))
            ras, decs, mags, catalogids = map(list, zip(*list(db_query.tuples())))
            ras = np.array(ras)
            decs = np.array(decs)
            mags = np.array(mags)
            db_query_results = (ras, decs, mags, catalogids)
    else:
        if instrument == 'BOSS':
            carts = ['ops_tycho2_brightneighbors',
                     'ops_gaia_brightneighbors']
            mag_col = targetdb.Magnitude.gaia_g
        else:
            carts = ['ops_2mass_psc_brightneighbors']
            mag_col = targetdb.Magnitude.h
        # run the query
        db_query = (targetdb.CartonToTarget.select(targetdb.Target.ra,
                                                   targetdb.Target.dec,
                                                   mag_col,
                                                   targetdb.CartonToTarget.pk)
                                           .join(targetdb.Target)
                                           .switch(targetdb.CartonToTarget)
                                           .join(targetdb.Magnitude)
                                           .switch(targetdb.CartonToTarget)
                                           .join(targetdb.Carton)
                                           .where((targetdb.Target.cone_search(racen,
                                                                               deccen,
                                                                               1.5)) &
                                                  (targetdb.Carton.carton.in_(carts))))
        ras, decs, mags, catalogids = map(list, zip(*list(db_query.tuples())))
        ras = np.array(ras)
        decs = np.array(decs)
        mags = np.array(mags)
        db_query_results = (ras, decs, mags, catalogids)
    return db_query_results


class DesignModeCheck(DesignMode):
    """
    Parameters
    ----------
    FPSDesign: object
        mugatu.fpsdesign.FPSDesign object with a built design.

    desmode_label: str
        The DesignMode label from targetdb. Options for label are:
        bright_time, dark_plane, dark_monit, dark_rm and dark_faint.

    desmode_manual: dict
        Dictonary of DesignMode parameters to be used as manual
        inputs to validate the design, rather than from targetdb.

    db_query_results_boss: dict
        Database query results for BOSS bright neighbor check.
        Each index of dict is a tuple of (ras, decs, mags, catalogids)
        with one index for designmode and the other safety.

    db_query_results_apogee: dict
        Database query results for APOGEE bright neighbor check.
        Each index of dict is a tuple of (ras, decs, mags, catalogids)
        with one index for designmode and the other safety.

    Attributes
    ----------
    design: dict
        FPSDesign dictonary.

    racen: float
        RA center of the field (degrees)

    deccen: float
        DEC center of the field (degrees)

    position_angle: float
        Position angle of the field E of N in degrees.

    observatory: str
        Observatory where observation is taking place, either
        'LCO' or 'APO'.

    obsTime: float
        Julian date of the observation.

    rg: kaiju.robotGrid
        Kaiju robotGrid object for the design.

    n_skies_min: dict
        Dictonary with the minimum number of skies for a design
        for each instrument ('APOGEE' and 'BOSS').

    min_skies_fovmetric: dict
        Dictonary wih the FOV metric for the skies in a design
        for each instrument('APOGEE' and 'BOSS').
        The FOV metric is described by three parameters, the
        nth neighbor to get distances to, the percentle distance
        to calculate and the distace to compare against for validation
        (in mm).

    n_stds_min: dict
        Dictonary with the minimum number of standards for a design
        for each instrument ('APOGEE' and 'BOSS').

    min_stds_fovmetric: dict
        Dictonary wih the FOV metric for the standards in a design
        for each instrument ('APOGEE' and 'BOSS').
        The FOV metric is described by three parameters, the
        nth neighbor to get distances to, the percentle distance
        to calculate and the distace to compare against for validation
        (in mm).

    stds_mags: dict
        Dictonary for the min/max magnitude for the standards in a
        design for each instrument ('APOGEE' and 'BOSS'). Indexes
        correspond to magntidues: [g, r, i, bp, gaia_g, rp, h].

    bright_limit_targets: dict
        Dictonary for the min/max magnitude for the science targets
        in adesign for each instrument ('APOGEE' and 'BOSS'). Indexes
        correspond to magntidues: [g, r, i, bp, gaia_g, rp, h].

    sky_neighbors_targets: dict
        Dictonary for the parameters used to check distance between
        skies and all possible sources in field for each instrument
        ('APOGEE' and 'BOSS'). Distances to targets (r, in arcseconds)
        must be r > R_0 * (lim - mags) ** beta, where mags is G band
        for BOSS and H band for APOGEE. Indexes correspond to:
        [R_0, beta, lim]

    trace_diff_targets: dict
        Dictonary for the maximum magnitude difference allowed between
        fibers next to each ther on the chip for each instrument
        ('APOGEE' and 'BOSS'). Here the magntidue difference is checked
        in the G band for BOSS and H band for APOGEE.

    carton_classes: dict
        Dictonary of arrays for the carton pks in each category
        (i.e. science, standards and skies).
    """

    def __init__(self, FPSDesign, desmode_label,
                 desmode_manual=None,
                 db_query_results_boss=None,
                 db_query_results_apogee=None):
        # grab needed info from FPSDesign object
        self.design = FPSDesign.design
        self.racen = FPSDesign.racen
        self.deccen = FPSDesign.deccen
        self.position_angle = FPSDesign.position_angle
        self.observatory = FPSDesign.observatory
        self.obsTime = FPSDesign.obsTime
        self.rg = FPSDesign.rg
        self.desmode_label = desmode_label
        self.db_query_results_boss = db_query_results_boss
        self.db_query_results_apogee = db_query_results_apogee

        # grab the design mode params
        if desmode_manual is None:
            self.fromdb(label=self.desmode_label)
        else:
            self.fromdict(designmode_dict=desmode_manual)

        # classify cartons as skies, standards or science
        self.carton_classes = {}
        self.carton_classes['science'] = []
        self.carton_classes['sky'] = []
        self.carton_classes['std'] = []
        for cat in np.unique(self.design['category'][self.design['catalogID'] != -1]):
            if 'sky' in cat:
                self.carton_classes['sky'] += list(self.design['category'][(self.design['catalogID'] != -1) &
                                                                            (self.design['category'] == cat)])
            elif 'standard' in cat:
                self.carton_classes['std'] += list(self.design['category'][(self.design['catalogID'] != -1) &
                                                                            (self.design['category'] == cat)])
            # I think else makes sense here as there are science
            # and open fiber labels?
            else:
                self.carton_classes['science'] += list(self.design['category'][(self.design['catalogID'] != -1) &
                                                                                (self.design['category'] == cat)])

        # collect magntiudes of design
        # here I am doing g,r,i,BP,G,RP,H
        self.mags = self.design['magnitudes']

    def skies_min(self, instrument, return_metric=False):
        """
        Checks if design has the required number of skies
        for some instrument. Returns True if number skies is
        greater than the minimum and False if not.

        Parameters
        ----------
        instrument: str
            Instrument to check number of sky fibers.
            Must be 'BOSS' or 'APOGEE'.

        return_metric: boolean
            If True, will return the number of skies
            in the design.

        Returns
        -------
        : boolean
            True if number skies is
            greater than the minimum and False if not.

        n_skies: int
            Number of skies in design, only returned if
            return_metric=True.
        """
        n_skies = len(self.design['catalogID'][(self.design['catalogID'] != -1) &
                                                     (np.isin(self.design['category'],
                                                              self.carton_classes['sky'])) &
                                                     (self.design['obsWavelength'] == instrument)])
        if return_metric:
            if n_skies >= self.n_skies_min[instrument]:
                return True, n_skies
            else:
                return False, n_skies
        else:
            if n_skies >= self.n_skies_min[instrument]:
                return True
            else:
                return False

    def stds_min(self, instrument, return_metric=False):
        """
        Checks if design has the required number of standards
        for some instrument. Returns True if number standards is
        greater than the minimum and False if not.

        Parameters
        ----------
        instrument: str
            Instrument to check number of standard fibers.
            Must be 'BOSS' or 'APOGEE'.

        return_metric: boolean
            If True, will return the number of standards
            in the design.

        Returns
        -------
        : boolean
            True if number standards is
            greater than the minimum and False if not.

        n_stds: int
            Number of standards in design, only returned if
            return_metric=True.
        """
        n_stds = len(self.design['catalogID'][(self.design['catalogID'] != -1) &
                                                    (np.isin(self.design['category'],
                                                             self.carton_classes['std'])) &
                                                    (self.design['obsWavelength'] == instrument)])
        if return_metric:
            if n_stds >= self.n_stds_min[instrument]:
                return True, n_stds
            else:
                return False, n_stds
        else:
            if n_stds >= self.n_stds_min[instrument]:
                return True
            else:
                return False

    def skies_fov(self, instrument, return_metric=False):
        """
        Checks if design meets the FOV metric for the skies
        for some instrument. Returns True FOV metric met,
        False if not. Also, if no science targets in design
        returns True, while no skies returns False.

        Parameters
        ----------
        instrument: str
            Instrument to FOV metric.
            Must be 'BOSS' or 'APOGEE'.

        return_metric: boolean
            If True, will return the FOV metric
            for the design.

        Returns
        -------
        : boolean
            True FOV metric met,
            False if not. Also, if no science targets in design
            returns True, while no skies returns False.

        perc_dist: float
            Percentile distance between science assignments and
            skies (FOV metric) for the design. Only returned if
            return_metric=True.
        """
        # for now, return True (passing) if no FOV metric supplied
        if np.all(self.min_skies_fovmetric[instrument] == -999.):
            if return_metric:
                return True, -1.
            else:
                return True
        # get x,y of the skies
        x_sky = self.design['x'][(self.design['catalogID'] != -1) &
                                 (np.isin(self.design['category'],
                                          self.carton_classes['sky'])) &
                                 (self.design['obsWavelength'] == instrument)]
        y_sky = self.design['y'][(self.design['catalogID'] != -1) &
                                 (np.isin(self.design['category'],
                                          self.carton_classes['sky'])) &
                                 (self.design['obsWavelength'] == instrument)]

        x_sci = self.design['x'][(self.design['catalogID'] != -1) &
                                 (~np.isin(self.design['category'],
                                           self.carton_classes['sky'])) &
                                 (self.design['obsWavelength'] == instrument)]
        y_sci = self.design['y'][(self.design['catalogID'] != -1) &
                                 (~np.isin(self.design['category'],
                                           self.carton_classes['sky'])) &
                                 (self.design['obsWavelength'] == instrument)]

        # if no science in band, doesnt matter?
        if len(x_sci) == 0:
            if return_metric:
                return True, -1.
            else:
                return True
        # if no skies required, dont do check
        if self.n_skies_min[instrument] == 0:
            if return_metric:
                return True, -1.
            else:
                return True
        # if no skies, dont do check
        if len(x_sky) == 0:
            if return_metric:
                return False, -1.
            else:
                return False

        # create KDE tree
        tree = cKDTree(np.column_stack((x_sky, y_sky)))
        # get distances for nearest neighbors
        dd, ii = tree.query(np.column_stack((x_sci, y_sci)),
                            k=self.min_skies_fovmetric[instrument][0])
        # second column is the nth neighbor distance if k>1
        if self.min_skies_fovmetric[instrument][0] == 1:
            dists = dd
        else:
            dists = dd[:, 1]
        # this assumes percentile is on 0 to 100 scale
        perc_dist = np.percentile(dists,
                                  self.min_skies_fovmetric[instrument][1])
        if return_metric:
            if perc_dist < self.min_skies_fovmetric[instrument][2]:
                return True, perc_dist
            else:
                return False, perc_dist
        else:
            if perc_dist < self.min_skies_fovmetric[instrument][2]:
                return True
            else:
                return False

    def stds_fov(self, instrument, return_metric=False):
        """
        Checks if design meets the FOV metric for the standards
        for some instrument. Returns True FOV metric met,
        False if not. Also, if no science targets in design
        returns True, while no standards returns False.

        Parameters
        ----------
        instrument: str
            Instrument to FOV metric.
            Must be 'BOSS' or 'APOGEE'.

        return_metric: boolean
            If True, will return the FOV metric
            for the design.

        Returns
        -------
        : boolean
            True FOV metric met,
            False if not. Also, if no science targets in design
            returns True, while no standards returns False.

        perc_dist: float
            Percentile distance between science assignments and
            standards (FOV metric) for the design. Only returned if
            return_metric=True.
        """
        # for now, return True (passing) if no FOV metric supplied
        if np.all(self.min_stds_fovmetric[instrument] == -999.):
            if return_metric:
                return True, -1.
            else:
                return True
        # get x,y of the standards
        x_std = self.design['x'][(self.design['catalogID'] != -1) &
                                 (np.isin(self.design['category'],
                                          self.carton_classes['std'])) &
                                 (self.design['obsWavelength'] == instrument)]
        y_std = self.design['y'][(self.design['catalogID'] != -1) &
                                 (np.isin(self.design['category'],
                                          self.carton_classes['std'])) &
                                 (self.design['obsWavelength'] == instrument)]

        x_sci = self.design['x'][(self.design['catalogID'] != -1) &
                                 (~np.isin(self.design['category'],
                                           self.carton_classes['std'])) &
                                 (self.design['obsWavelength'] == instrument)]
        y_sci = self.design['y'][(self.design['catalogID'] != -1) &
                                 (~np.isin(self.design['category'],
                                           self.carton_classes['std'])) &
                                 (self.design['obsWavelength'] == instrument)]

        # if no science in band, doesnt matter?
        if len(x_sci) == 0:
            if return_metric:
                return True, -1.
            else:
                return True
        # if no stds required, dont do check
        if self.n_stds_min[instrument] == 0:
            if return_metric:
                return True, -1.
            else:
                return True
        # if no stds, dont do check
        if len(x_std) == 0:
            if return_metric:
                return False, -1.
            else:
                return False

        # create KDE tree
        tree = cKDTree(np.column_stack((x_std, y_std)))
        # get distances for nearest neighbors
        dd, ii = tree.query(np.column_stack((x_sci, y_sci)),
                            k=self.min_stds_fovmetric[instrument][0])
        # second column is the nth neighbor distance if k>1
        if self.min_stds_fovmetric[instrument][0] == 1:
            dists = dd
        else:
            dists = dd[:, 1]
        # this assumes percentile is on 0 to 100 scale
        perc_dist = np.percentile(dists,
                                  self.min_stds_fovmetric[instrument][1])
        if return_metric:
            if perc_dist < self.min_stds_fovmetric[instrument][2]:
                return True, perc_dist
            else:
                return False, perc_dist
        else:
            if perc_dist < self.min_stds_fovmetric[instrument][2]:
                return True
            else:
                return False

    def mag_limits(self, mag_metric,
                   instrument, carton_class):
        """
        Checks the if magnitude of assignments agree with
        design mode for some instrument and carton class

        Parameters
        ----------
        mag_metric: np.array
            Array of shape (N,M), where N=7 corresponds to
            magntiudes [g, r, i, bp, gaia_g, rp, h], and M=2
            where 0th column is minimum magnitude and 1st column
            is maximum magnitude. If no check in certain band,
            use None as value.

        instrument: str
            Instrument to check magnitudes.
            Must be 'BOSS' or 'APOGEE'.

        carton_class: str
            Carton class to check magnitude limits of.
            Must be 'std' or 'science'.

        Returns
        -------
        mag_checks: np.array
            Array of booleans equal to length of self.design.
            If True, assignment within magnitude limits, False
            if not. If assignment is not with instrument or in
            carton_class, will be False.

        complete_check: np.array
            Array of str equal to length of self.design.
            If 'INCOMPLETE', then assignment either did not have
            all required photometry for check, or assignment is
            not with instrument or in carton_class.
        """
        mag_checks = np.zeros(len(self.design['catalogID']),
                              dtype=bool)
        # if complete, all bands for target present in check
        # if incomplete, then
        complete_check = np.array(['COMPLETE' for _ in range(len(self.design['catalogID']))],
                                  dtype='<U12')

        # check which limits are defined for mode
        check_inds = []
        for i in range(mag_metric.shape[0]):
            if mag_metric[i][0] != -999. or mag_metric[i][1] != -999.:
                check_inds.append(i)

        # run checks
        for i in range(len(mag_checks)):
            if (self.design['catalogID'][i] != -1 and
                self.design['category'][i] in self.carton_classes[carton_class] and
                self.design['obsWavelength'][i] == instrument):
                # check in each band that has check defined
                targ_check = np.zeros(len(check_inds), dtype=bool)
                for j, ind in enumerate(check_inds):
                    # check the magntiude for this assignment
                    targ_check[j], complete_check[i] = check_assign_mag_limit(
                                                            mag_metric[ind][0],
                                                            mag_metric[ind][1],
                                                            self.mags[i][ind])
                # if all True, then passes
                if np.all(targ_check):
                    mag_checks[i] = True
            else:
                complete_check[i] = 'INCOMPLETE'
        return mag_checks, complete_check

    def bright_neighbors(self, instrument, check_type='designmode',
                         db_query=None):
        """
        Check if any fibers are placed near a bright star

        Parameters
        ----------
        instrument: str
            Instrument to check sky distances.
            Must be 'BOSS' or 'APOGEE'.

        check_type: str
            Either 'designmode' for checking bright
            stars in catalogdb down to designmode magnitude
            limit, or 'safety' for checking bright stars in
            targetdb in bright star cartons.

        db_query: peewee query
            Optionally pass the pre-queried bright stars
            from catalogdb or targetdb. Must include ra, dec,
            magnitude and id. Other option is to provide tuple
            of (ra,dec,magntiude,id).

        Returns
        -------
        neigh_checks_des: np.array
            Array of booleans equal to length of 500, where order
            is robotID 1 to 500 robotID.
            True assignment valid (i.e. not near bright star),
            False if not valid.

        hasFiber: np.array
            Array of booleans equal to length of 500, where order
            is robotID 1 to 500 robotID. True if robotID has intrument
            on fiber, False if not. If False, do not consider result from
            neigh_checks_des for this robotID

        mag_adj_robo: np.array
            If neigh_checks_des == False, reports the magntiude of nearby
            bright source at the fiber position. If neigh_checks_des == True,
            then no value reported (where NULL = -9999.).

        isassigned: np.array
            Array of booleans equal to length of 500, where order
            is robotID 1 to 500 robotID. True if robotID is assigned in grid
            and False if it is not.
        """
        # get xPos and yPos from robotGrid
        xrobo = np.zeros(500)
        yrobo = np.zeros(500)
        hasFiber = np.zeros(500, dtype=bool) + True
        isassigned = np.zeros(500, dtype=bool) + True
        mag_adj_robo = np.zeros(500) - 9999.
        for i, robotID in enumerate(self.rg.robotDict):
            if instrument == 'BOSS':
                xrobo[i] = self.rg.robotDict[robotID].bossWokXYZ[0]
                yrobo[i] = self.rg.robotDict[robotID].bossWokXYZ[1]
            else:
                hasFiber[i] = self.rg.robotDict[robotID].hasApogee
                xrobo[i] = self.rg.robotDict[robotID].apWokXYZ[0]
                yrobo[i] = self.rg.robotDict[robotID].apWokXYZ[1]
            isassigned[i] = self.rg.robotDict[robotID].isAssigned()

        ra_robo, dec_robo, fieldWarn = wokxy2radec(xWok=xrobo,
                                                   yWok=yrobo,
                                                   waveName=instrument.title(),
                                                   raCen=self.racen,
                                                   decCen=self.deccen,
                                                   obsAngle=self.position_angle,
                                                   obsSite=self.observatory,
                                                   obsTime=self.obsTime)

        neigh_checks = np.zeros(500, dtype=bool) + True
        # run query for field if not supplied
        if instrument == 'BOSS':
             # grab r_sdss limit for boss
            if 'bright' in self.desmode_label:
                # no r_sdss for bright so do g band
                # this is hacky and needs to be fixed!!!
                mag_lim = self.bright_limit_targets['BOSS'][0][0]
            else:
                mag_lim = self.bright_limit_targets['BOSS'][1][0]
        else:
            # grab h 2mass mag for limit
            mag_lim = self.bright_limit_targets['APOGEE'][6][0]
        if (self.db_query_results_boss is not None and
           instrument == 'BOSS'):
            db_query = self.db_query_results_boss[check_type]
        elif (self.db_query_results_apogee is not None and
           instrument == 'APOGEE'):
            db_query = self.db_query_results_apogee[check_type]
        elif db_query is None:
            db_query = build_brigh_neigh_query(check_type, instrument,
                                               mag_lim, self.racen,
                                               self.deccen)

        # only do check if any stars returned
        if len(db_query) > 0:
            if isinstance(db_query, tuple):
                ras, decs, mags, catalogids = db_query
            else:
                ras, decs, mags, catalogids = map(list, zip(*list(db_query.tuples())))

            if 'bright' in self.desmode_label:
                r_exclude = bright_neigh_exclusion_r(mags,
                                                     mag_lim,
                                                     lunation='bright')
            else:
                r_exclude = bright_neigh_exclusion_r(mags,
                                                     mag_lim,
                                                     lunation='dark')

            # check if fibers too close to bright neighbors
            for i in range(len(r_exclude)):
                # only do check if exclusion radius larger
                # than 0" (otherwise star not bright enough)
                if r_exclude[i] > 0.:
                    dist = ang_sep(ras[i], decs[i],
                                   ra_robo, dec_robo) * 3600.
                    if 'bright' in self.desmode_label:
                        mag_adj = adjusted_brigh_neigh_mag(mags[i],
                                                           dist,
                                                           lunation='bright')
                    else:
                        mag_adj = adjusted_brigh_neigh_mag(mags[i],
                                                           dist,
                                                           lunation='dark')
                    neigh_checks[dist < r_exclude[i]] = False
                    mag_adj_robo[dist < r_exclude[i]] = mag_adj[dist < r_exclude[i]]
        # dont accoutn for places where no fiber
        mag_adj_robo[~hasFiber] = -9999.

        return neigh_checks, hasFiber, mag_adj_robo, isassigned

    def design_mode_check_all(self, verbose=True):
        """
        Perform all DesignMode checks for the given design.

        Parameters
        ----------
        verbose: boolean
            True if want print statement summarizing results
            of all checks.

        Attributes
        ----------
        n_skies_min_check: dict
            Results of minumum sky check for each instrument.

        min_skies_fovmetric_check: dict
            Results of sky FOV metric check for each instrument.

        n_stds_min_check: dict
            Results of minumum standard check for each instrument.

        min_stds_fovmetric_check: dict
            Results of standard FOV metric check for each instrument.

        stds_mags_check: dict
            Results of standard magnitude limit check for
            each instrument.

        bright_limit_targets_check: dict
             Results of science magnitude limit check for
            each instrument.
        """
        self.n_skies_min_check = {}
        result = self.skies_min(instrument='BOSS',
                                return_metric=True)
        self.n_skies_min_check['BOSS'] = result[0]
        self.n_skies_min_check['BOSS_metric'] = result[1]
        result = self.skies_min(instrument='APOGEE',
                                return_metric=True)
        self.n_skies_min_check['APOGEE'] = result[0]
        self.n_skies_min_check['APOGEE_metric'] = result[1]

        self.min_skies_fovmetric_check = {}
        result = self.skies_fov(instrument='BOSS',
                                return_metric=True)
        self.min_skies_fovmetric_check['BOSS'] = result[0]
        self.min_skies_fovmetric_check['BOSS_metric'] = result[1]
        result = self.skies_fov(instrument='APOGEE',
                                return_metric=True)
        self.min_skies_fovmetric_check['APOGEE'] = result[0]
        self.min_skies_fovmetric_check['APOGEE_metric'] = result[1]

        self.n_stds_min_check = {}
        result = self.stds_min(instrument='BOSS',
                               return_metric=True)
        self.n_stds_min_check['BOSS'] = result[0]
        self.n_stds_min_check['BOSS_metric'] = result[1]
        result = self.stds_min(instrument='APOGEE',
                               return_metric=True)
        self.n_stds_min_check['APOGEE'] = result[0]
        self.n_stds_min_check['APOGEE_metric'] = result[1]

        self.min_stds_fovmetric_check = {}
        result = self.stds_fov(instrument='BOSS',
                               return_metric=True)
        self.min_stds_fovmetric_check['BOSS'] = result[0]
        self.min_stds_fovmetric_check['BOSS_metric'] = result[1]
        result = self.stds_fov(instrument='APOGEE',
                               return_metric=True)
        self.min_stds_fovmetric_check['APOGEE'] = result[0]
        self.min_stds_fovmetric_check['APOGEE_metric'] = result[1]

        self.stds_mags_check = {}
        self.stds_mags_check['BOSS'] = self.mag_limits(self.stds_mags['BOSS'],
                                                       'BOSS',
                                                       'std')
        check_tot = len(self.stds_mags_check['BOSS'][0][self.stds_mags_check['BOSS'][0]])
        design_tot = len(self.design['x'][(self.design['catalogID'] != -1) &
                                          (np.isin(self.design['category'],
                                                   self.carton_classes['std'])) &
                                          (self.design['obsWavelength'] == 'BOSS')])
        self.stds_mags_check['BOSS_metric'] = [check_tot, design_tot]
        self.stds_mags_check['APOGEE'] = self.mag_limits(self.stds_mags['APOGEE'],
                                                         'APOGEE',
                                                         'std')
        check_tot = len(self.stds_mags_check['APOGEE'][0][self.stds_mags_check['APOGEE'][0]])
        design_tot = len(self.design['x'][(self.design['catalogID'] != -1) &
                                          (np.isin(self.design['category'],
                                                   self.carton_classes['std'])) &
                                          (self.design['obsWavelength'] == 'APOGEE')])
        self.stds_mags_check['APOGEE_metric'] = [check_tot, design_tot]

        self.bright_limit_targets_check = {}
        self.bright_limit_targets_check['BOSS'] = self.mag_limits(self.bright_limit_targets['BOSS'],
                                                                  'BOSS',
                                                                  'science')
        check_tot = len(self.bright_limit_targets_check['BOSS'][0][self.bright_limit_targets_check['BOSS'][0]])
        design_tot = len(self.design['x'][(self.design['catalogID'] != -1) &
                                          (np.isin(self.design['category'],
                                                   self.carton_classes['science'])) &
                                          (self.design['obsWavelength'] == 'BOSS')])
        self.bright_limit_targets_check['BOSS_metric'] = [check_tot, design_tot]
        self.bright_limit_targets_check['APOGEE'] = self.mag_limits(self.bright_limit_targets['APOGEE'],
                                                                    'APOGEE',
                                                                    'science')
        check_tot = len(self.bright_limit_targets_check['APOGEE'][0][self.bright_limit_targets_check['APOGEE'][0]])
        design_tot = len(self.design['x'][(self.design['catalogID'] != -1) &
                                          (np.isin(self.design['category'],
                                                   self.carton_classes['science'])) &
                                          (self.design['obsWavelength'] == 'APOGEE')])
        self.bright_limit_targets_check['APOGEE_metric'] = [check_tot, design_tot]

        self.bright_neighbor_check = {}
        self.bright_neighbor_check['BOSS'] = self.bright_neighbors(instrument='BOSS',
                                                                   check_type='designmode')
        check_tot = len(self.bright_neighbor_check['BOSS'][0][self.bright_neighbor_check['BOSS'][0] &
                                                              self.bright_neighbor_check['BOSS'][1] &
                                                              self.bright_neighbor_check['BOSS'][3]])
        design_tot = len(self.bright_neighbor_check['BOSS'][0][self.bright_neighbor_check['BOSS'][1] &
                                                               self.bright_neighbor_check['BOSS'][3]])
        mag_adj_near_bs = self.bright_neighbor_check['BOSS'][2]
        self.bright_neighbor_check['BOSS_metric'] = [check_tot, design_tot, mag_adj_near_bs]
        self.bright_neighbor_check['APOGEE'] = self.bright_neighbors(instrument='APOGEE',
                                                                     check_type='designmode')
        check_tot = len(self.bright_neighbor_check['APOGEE'][0][self.bright_neighbor_check['APOGEE'][0] &
                                                                self.bright_neighbor_check['APOGEE'][1] &
                                                                self.bright_neighbor_check['APOGEE'][3]])
        design_tot = len(self.bright_neighbor_check['APOGEE'][0][self.bright_neighbor_check['APOGEE'][1] &
                                                                 self.bright_neighbor_check['APOGEE'][3]])
        mag_adj_near_bs = self.bright_neighbor_check['APOGEE'][2]
        self.bright_neighbor_check['APOGEE_metric'] = [check_tot, design_tot, mag_adj_near_bs]

        if verbose:
            verbose_output = ''
            verbose_output += 'DesignMode Param                  | Pass Check?\n'
            verbose_output += '----------------------------------|-----------------\n'
            verbose_output += 'N Skies Min (BOSS):               | %s\n' % self.n_skies_min_check['BOSS']
            verbose_output += 'N Skies Min (APOGEE):             | %s\n' % self.n_skies_min_check['APOGEE']
            verbose_output += 'N Standards Min (BOSS):           | %s\n' % self.n_stds_min_check['BOSS']
            verbose_output += 'N Standards Min (APOGEE):         | %s\n' % self.n_stds_min_check['APOGEE']
            verbose_output += 'FOV Metric Skies (BOSS):          | %s\n' % self.min_skies_fovmetric_check['BOSS']
            verbose_output += 'FOV Metric Skies (APOGEE):        | %s\n' % self.min_skies_fovmetric_check['APOGEE']
            verbose_output += 'FOV Metric Standards (BOSS):      | %s\n' % self.min_stds_fovmetric_check['BOSS']
            verbose_output += 'FOV Metric Standards (APOGEE):    | %s\n' % self.min_stds_fovmetric_check['APOGEE']

            check_tot = self.stds_mags_check['BOSS_metric'][0]
            design_tot = self.stds_mags_check['BOSS_metric'][1]
            verbose_output += 'Magnitude Limit Stds (BOSS):      | %d out of %d\n' % (check_tot, design_tot)

            check_tot = self.stds_mags_check['APOGEE_metric'][0]
            design_tot = self.stds_mags_check['APOGEE_metric'][1]
            verbose_output += 'Magnitude Limit Stds (APOGEE):    | %d out of %d\n' % (check_tot, design_tot)

            check_tot = self.bright_limit_targets_check['BOSS_metric'][0]
            design_tot = self.bright_limit_targets_check['BOSS_metric'][1]
            verbose_output += 'Magnitude Limit Targets (BOSS):   | %d out of %d\n' % (check_tot, design_tot)

            check_tot = self.bright_limit_targets_check['APOGEE_metric'][0]
            design_tot = self.bright_limit_targets_check['APOGEE_metric'][1]
            verbose_output += 'Magnitude Limit Targets (APOGEE): | %d out of %d\n' % (check_tot, design_tot)

            check_tot = self.bright_neighbor_check['BOSS_metric'][0]
            design_tot = self.bright_neighbor_check['BOSS_metric'][1]
            verbose_output += 'Bright Neighbor Check (BOSS):     | %d out of %d\n' % (check_tot, design_tot)

            check_tot = self.bright_neighbor_check['APOGEE_metric'][0]
            design_tot = self.bright_neighbor_check['APOGEE_metric'][1]
            verbose_output += 'Bright Neighbor Check (APOGEE):   | %d out of %d\n' % (check_tot, design_tot)

            print(verbose_output)



# @Author: Ilija Medan
# @Date: December 29, 2020
# @Filename: fpsdesign.py
# @License: BSD 3-Clause
# @Copyright: Ilija Medan

import numpy as np
import warnings
from astropy.io import fits
from astropy.time import Time
from peewee import JOIN

import kaiju
import kaiju.robotGrid
from mugatu.exceptions import (MugatuError, MugatuWarning,
                               MugatuDesignError, MugatuDesignWarning,
                               MugatuDesignModeWarning)
from coordio.utils import radec2wokxy
from coordio.defaults import POSITIONER_HEIGHT
from mugatu.designs_to_targetdb import (make_design_assignments_targetdb,
                                        make_design_field_targetdb)
from mugatu.designmode import DesignModeCheck

try:
    from sdssdb.peewee.sdss5db import database
    database.set_profile('operations')
    _database = True
except:
    _database = False

from sdssdb.peewee.sdss5db.targetdb import (Design, Field, Observatory,
                                            Assignment, Instrument, Target,
                                            Hole, CartonToTarget, Carton,
                                            Magnitude, Category)


class FPSDesign(object):
    """
    Class to load, validate and export a design.

    Parameters
    ----------
    design_pk: int
        The pk of the design as it appears in targetdb.

    obsTime: float
        Julian date of the observation.

    racen: np.float64
        Right Ascension of center of the field in degrees.

    deccen: np.float64
        Declination of center of the field in degrees.

    position_angle: np.float64
        Position angle of the field E of N in degrees.

    observatory: str
        Observatory where observation is taking place, either
        'LCO' or 'APO'.

    desmode_label: int
        The pk in targetdb for the observing mode for the design.

    idtype: str
        ID type used for catalogids in design. Must be 'catalogID'
        or 'carton_to_target'.

    catalogids: np.array
        List of ids for a manual design in db. Default from idtype
        is carton_to_target.

    ra: np.array
        List of right ascensions that correspond to catalogids.
        If list not provided, but catalogids for a manual design
        are provided, right ascensions will be pulled from
        targetdb.

    dec: np.array
        List of declinations that correspond to catalogids.
        If list not provided, but catalogids for a manual design
        are provided, declinations will be pulled from
        targetdb.

    pmra: np.array
        Array of proper motions in the RA axis for the targets, in
        milliarcsec/yr. Must be a true angle, i.e, it must include
        the cos(dec) term. If no proper motion, set index to zero.

    pmdec: np.array
        Array of proper motions in the DEC axis for the targets, in
        milliarcsec/yr. If no proper motion, set index to zero.

    delta_ra: np.array
        List of offsets in right ascensions that correspond to catalogids.
        If list not provided, but catalogids for a manual design
        are provided, right ascensions will be pulled from
        targetdb.

    delta_dec: np.array
        List of offsets in declinations that correspond to catalogids.
        If list not provided, but catalogids for a manual design
        are provided, declinations will be pulled from
        targetdb.

    epoch: np.array
        Array of epochs for the coordinates in decimal years.

    holeID: np.array
        Fiber assignement for each target in the
        manual design.

    obsWavelength: np.array
        Wavelength of observation for each fiber
        that correspond to catalogids.
        Must be either 'BOSS' or 'APOGEE'.

    priority: np.array
        Priorties for targets that correspond to catalogids.

    carton_pk: np.array
        The carton_pks (as in targetdb) for the targets
        that correspond to catalogids.

    category: np.array
        The category for each target. Can be 'science', 'sky_INSTRUMENT'
        or 'standard_INSTRUMENT'

    magnitudes: np.array
        Magnitudes of the targets. Should be of size (N, 7), where
        columns correspond to g, r, i, bp, gaia_g, rp and h band
        magnitudes.

    design_file: str
        FITS file with a manual design. The file must be the same
        format as the rsFieldAssignments file.

    manual_design: boolean
        Boolean if the design being validated is manual
        (manual_design=True) or in targetdb (manual_design=False)

    exp: int
        Exposure number for design in file. If exp=0, assumes
        only 1 exposure in the design file. if exp>0, then will
        choose exposure exp = 1 to N.

    Attributes
    ----------
    design: dict
        contains all of the design inputs for Kaiju needed to validate
        a design.

    rg: Kaiju RobotGrid object
        Kaiju RobotGrid with the assignments from design.

    targets_unassigned: list
        catalogid of targets that could not be assigned due to assigned
        robot not being able to reach the assigned target.

    holeID_mapping: np.array
        Mapping between robotID and holeID. Index corresponds to robotID
        in robotID_mapping attrobute.

    robotID_mapping: np.array
        Mapping between robotID and holeID. Index corresponds to holeID
        in holeID_mapping attrobute.

    targets_collided: list
        catalogid of targets that could not be assigned due to assigned
        robot to assigned target resulting in a collision.

    valid_design: dictonary
        Same format as design dictonary, except this is validated
        design by Kaiju with collision/delock assignments removed.

    design_built: booleen
        True if design dictonary has been populated, False if not.

    hourAngle: float
        Hour angle of field center from coordio.utils.radec2wokxy.

    positionAngle_coordio: float
        position angle of field center from coordio.utils.radec2wokxy.

    design_errors: dict
        dictonary with errors for design. Created and filled after validation.
    """

    def __init__(self, design_pk, obsTime, racen=None, deccen=None,
                 position_angle=None, observatory=None, desmode_label=None,
                 idtype='carton_to_target', catalogids=None, ra=None, dec=None,
                 pmra=None, pmdec=None, delta_ra=None, delta_dec=None,
                 epoch=None, holeID=None, obsWavelength=None,
                 priority=None, carton_pk=None, category=None, magnitudes=None,
                 design_file=None, manual_design=False, exp=0):
        if idtype != 'catalogID' and idtype != 'carton_to_target':
            message = 'idtype must be catalogID or carton_to_target'
            raise MugatuError(message=message)
        self.design_pk = design_pk
        self.obsTime = obsTime
        self.design = {}
        # either set field params or pull from db is design
        # is in targetdb
        if manual_design:
            if design_file is None:
                self.racen = racen
                self.deccen = deccen
                self.position_angle = position_angle
                self.observatory = observatory
                self.desmode_label = desmode_label
            else:
                head = fits.open(design_file)[0].header
                desmode_labels = head['DESMODE'].split(' ')
                self.racen = head['RACEN']
                self.deccen = head['DECCEN']
                self.position_angle = head['PA']
                self.observatory = head['obs'].strip().upper()
                self.desmode_label = desmode_labels[exp - 1]
        else:
            design_field_db = (
                Design.select(Design.design_id,
                              Design.design_mode,
                              Field.racen,
                              Field.deccen,
                              Field.position_angle,
                              Observatory.label)
                      .join(Field,
                            on=(Design.field_pk == Field.pk))
                      .join(Observatory,
                            on=(Field.observatory == Observatory.pk))
                      .where(Design.design_id == self.design_pk))

            self.racen = design_field_db.objects()[0].racen
            self.deccen = design_field_db.objects()[0].deccen
            self.position_angle = design_field_db.objects()[0].position_angle
            self.observatory = design_field_db.objects()[0].label
            self.desmode_label = design_field_db.objects()[0].design_mode.label

        # should these be catalogids or carton_to_target?
        self.idtype = idtype
        self.catalogids = catalogids
        self.ra = ra
        self.dec = dec
        self.pmra = pmra
        self.pmdec = pmdec
        self.delta_ra = delta_ra
        self.delta_dec = delta_dec
        self.epoch = epoch
        self.holeID = holeID
        self.obsWavelength = obsWavelength
        self.priority = priority
        self.carton_pk = carton_pk
        self.category = category
        self.magnitudes = magnitudes
        self.design_file = design_file
        self.manual_design = manual_design
        self.exp = exp

        # set dummy value for collision for now
        # this may want to be a input, not sure the standard here
        # initialize robotGrid
        # check wth Mike about stepsize?
        if self.observatory == 'APO':
            self.rg = kaiju.robotGrid.RobotGridAPO(stepSize=0.05)
        else:
            self.rg = kaiju.robotGrid.RobotGridLCO(stepSize=0.05)
        self.holeID_mapping = np.zeros(500, dtype='<U10')
        self.robotID_mapping = np.zeros(500, dtype=int)
        for i, robotID in enumerate(self.rg.robotDict):
            self.holeID_mapping[i] = self.rg.robotDict[robotID].holeID
            self.robotID_mapping[i] = robotID
        # this is in Conor's test, I'm not quite sure what it does
        # but without paths wont generate
        for k in self.rg.robotDict.keys():
            self.rg.homeRobot(k)
        # for rID in self.rg.robotDict:
            # robot = self.rg.getRobot(rID)
            # robot.setXYUniform()
        self.targets_unassigned = []
        self.targets_collided = []

        self.valid_design = {}

        self.design_built = False

    def _offset_radec(self, ra=None, dec=None, delta_ra=0., delta_dec=0.):
        """Offsets ra and dec according to specified amount. From Mike's
        robostrategy.Field object

        Parameters
        ----------
        ra : np.float64 or ndarray of np.float64
        right ascension, deg
        dec : np.float64 or ndarray of np.float64
            declination, deg
        delta_ra : np.float64 or ndarray of np.float64
            right ascension direction offset, arcsec
        delta_dec : np.float64 or ndarray of np.float64
            declination direction offset, arcsec

        Returns
        -------
        offset_ra : np.float64 or ndarray of np.float64
            offset right ascension, deg
        offset_dec : np.float64 or ndarray of np.float64
            offset declination, deg

        Notes
        -----
        Assumes that delta_ra, delta_dec are in proper coordinates; i.e.
        an offset of delta_ra=1 arcsec represents the same angular separation
        on the sky at any declination.
        Carefully offsets in the local directions of ra, dec based on
        the local tangent plane (i.e. does not just scale delta_ra by
        1/cos(dec))
        """
        deg2rad = np.pi / 180.
        arcsec2rad = np.pi / 180. / 3600.
        x = np.cos(dec * deg2rad) * np.cos(ra * deg2rad)
        y = np.cos(dec * deg2rad) * np.sin(ra * deg2rad)
        z = np.sin(dec * deg2rad)
        ra_x = - np.sin(ra * deg2rad)
        ra_y = np.cos(ra * deg2rad)
        ra_z = 0.
        dec_x = - np.sin(dec * deg2rad) * np.cos(ra * deg2rad)
        dec_y = - np.sin(dec * deg2rad) * np.sin(ra * deg2rad)
        dec_z = np.cos(dec * deg2rad)
        xoff = x + (ra_x * delta_ra + dec_x * delta_dec) * arcsec2rad
        yoff = y + (ra_y * delta_ra + dec_y * delta_dec) * arcsec2rad
        zoff = z + (ra_z * delta_ra + dec_z * delta_dec) * arcsec2rad
        offnorm = np.sqrt(xoff**2 + yoff**2 + zoff**2)
        xoff = xoff / offnorm
        yoff = yoff / offnorm
        zoff = zoff / offnorm
        decoff = np.arcsin(zoff) / deg2rad
        raoff = ((np.arctan2(yoff, xoff) / deg2rad) + 360.) % 360.
        return(raoff, decoff)

    def radec_to_xy(self, ev):
        """
        transorm radec to wok xy

        Parameters
        ---------
        ev: eval
            eval term to ignore unassigned fibers

        Returns
        -------
        fieldWarn: boolean
            Warning if any of the transformations
            should be eyed with suspicion.
        """
        with warnings.catch_warnings(record=True) as w:
            radVel = (np.zeros(len(self.design['ra_off'][ev]),
                               dtype=np.float64) + 1.e-4)
            parallax = (np.zeros(len(self.design['ra_off'][ev]),
                                 dtype=np.float64) + 1.e-4)
            res = radec2wokxy(
                ra=self.design['ra_off'][ev],
                dec=self.design['dec_off'][ev],
                coordEpoch=Time(self.design['epoch'][ev],
                                format='decimalyear').jd,
                waveName=np.array(list(map(lambda x: x.title(),
                                           self.design['obsWavelength'][ev]))),
                raCen=self.racen,
                decCen=self.deccen,
                obsAngle=self.position_angle,
                obsSite=self.observatory,
                obsTime=self.obsTime,
                pmra=self.design['pmra'][ev],
                pmdec=self.design['pmdec'][ev],
                parallax=parallax,
                radVel=radVel)
            self.design['x'][ev] = res[0]
            self.design['y'][ev] = res[1]
            fieldWarn = res[2]
            self.hourAngle = res[3]
            self.positionAngle_coordio = res[4]
            if len(w) > 0:
                for wm in w:
                    if 'iauPmsafe return' in wm.message.args[0]:
                        flag = ('Coordio xy coordinates converted '
                                'should be eyed with suspicion.')
                        warnings.warn(flag, MugatuDesignWarning)
        return fieldWarn

    def holeID_to_robotID(self, holeID, return_ind=True):
        """
        return the robotID for a given holeID based on current
        kaiju mapping
        """
        ind = np.where(self.holeID_mapping == holeID)[0][0]
        robotID = self.robotID_mapping[ind]
        if return_ind:
            return robotID, ind
        else:
            return robotID

    def build_design_db(self):
        """
        Populate the design dictonary for design in targetdb.

        Notes
        -----
        This method will build a design from the db if manual_design=False

        """

        if self.manual_design:
            flag = 'Building database design with manual_design=True flag'
            warnings.warn(flag, MugatuWarning)

        # initialize dict for the design
        # not using None or nan for no assignments
        # using -1 (for int) and -9999.99 (for float) for None assignment
        self.design['design_pk'] = self.design_pk
        self.design['desmode_label'] = self.desmode_label
        self.design['catalogID'] = np.zeros(500, dtype=np.int64) - 1
        self.design['robotID'] = np.zeros(500, dtype=np.int64) - 1
        self.design['holeID'] = np.zeros(500, dtype='<U10')
        self.design['obsWavelength'] = np.zeros(500, dtype='<U6')
        self.design['priority'] = np.zeros(500, dtype=int) - 1
        self.design['carton_pk'] = np.zeros(500, dtype=int) - 1
        self.design['category'] = np.zeros(500, dtype='<U10')
        self.design['ra'] = np.zeros(500, dtype=float) - 9999.99
        self.design['dec'] = np.zeros(500, dtype=float) - 9999.99
        self.design['delta_ra'] = np.zeros(500, dtype=float) - 9999.99
        self.design['delta_dec'] = np.zeros(500, dtype=float) - 9999.99
        self.design['epoch'] = np.zeros(500, dtype=float) - 9999.99
        self.design['ra_off'] = np.zeros(500, dtype=float) - 9999.99
        self.design['dec_off'] = np.zeros(500, dtype=float) - 9999.99
        self.design['pmra'] = np.zeros(500, dtype=float) - 9999.99
        self.design['pmdec'] = np.zeros(500, dtype=float) - 9999.99
        self.design['x'] = np.zeros(500, dtype=float) - 9999.99
        self.design['y'] = np.zeros(500, dtype=float) - 9999.99
        self.design['magnitudes'] = np.zeros((500, 7), dtype=float) - 9999.99

        # need to add wokHole to query when in db (not there now)
        # I need to test this when v05 is up, im unsure about Joins
        design_targ_db = (
            Assignment.select(Target.catalogid,
                              Hole.holeid,
                              Instrument.label,
                              CartonToTarget.priority,
                              Target.ra,
                              Target.dec,
                              CartonToTarget.carton,
                              Carton.pk.alias('carton_pk'),
                              CartonToTarget.pk,
                              Magnitude.g,
                              Magnitude.r,
                              Magnitude.i,
                              Magnitude.bp,
                              Magnitude.gaia_g,
                              Magnitude.rp,
                              Magnitude.h,
                              CartonToTarget.delta_ra,
                              CartonToTarget.delta_dec,
                              Target.pmra,
                              Target.pmdec,
                              Target.epoch,
                              Category.label.alias('cat_lab'))
                      .join(Hole)
                      .switch(Assignment)
                      .join(Instrument)
                      .switch(Assignment)
                      .join(CartonToTarget)
                      .join(Target)
                      .switch(CartonToTarget)
                      .join(Magnitude, JOIN.LEFT_OUTER)
                      .switch(CartonToTarget)
                      .join(Carton, JOIN.LEFT_OUTER)
                      .join(Category, JOIN.LEFT_OUTER)
                      .where(Assignment.design_id == self.design_pk))

        for d in design_targ_db.objects():
            # assign to index that corresponds to fiber assignment
            # index should match length of arrays
            robotID, pos_ind = self.holeID_to_robotID(d.holeid)
            if self.idtype == 'carton_to_target':
                self.design['catalogID'][pos_ind] = d.pk
            else:
                self.design['catalogID'][pos_ind] = d.catalogid

            self.design['robotID'][pos_ind] = robotID
            self.design['holeID'][pos_ind] = d.holeid
            # design['wokHoleID'][i] = design_targ_db[i]
            self.design['obsWavelength'][pos_ind] = d.label
            # catch targets with no assigned priority
            try:
                self.design['priority'][pos_ind] = d.priority
            except AttributeError:
                self.design['priority'][pos_ind] = -1
            self.design['carton_pk'][pos_ind] = d.carton_pk
            self.design['category'][pos_ind] = d.cat_lab
            self.design['ra'][pos_ind] = d.ra
            self.design['dec'][pos_ind] = d.dec
            self.design['delta_ra'][pos_ind] = d.delta_ra
            self.design['delta_dec'][pos_ind] = d.delta_dec
            self.design['pmra'][pos_ind] = d.pmra
            self.design['pmdec'][pos_ind] = d.pmdec
            self.design['epoch'][pos_ind] = d.epoch
            self.design['magnitudes'][pos_ind][0] = d.g
            self.design['magnitudes'][pos_ind][1] = d.r
            self.design['magnitudes'][pos_ind][2] = d.i
            self.design['magnitudes'][pos_ind][3] = d.bp
            self.design['magnitudes'][pos_ind][4] = d.gaia_g
            self.design['magnitudes'][pos_ind][5] = d.rp
            self.design['magnitudes'][pos_ind][6] = d.h

        # set nan pm tp zero
        self.design['pmra'][np.isnan(self.design['pmra'])] = 0.
        self.design['pmdec'][np.isnan(self.design['pmdec'])] = 0.
        # here convert ra/dec to x/y based on field/time of observation
        # I think I need to add inertial in here at some point,
        # dont see this in targetdb though
        ev = eval("(self.design['ra'] != -9999.99)")
        res = self._offset_radec(ra=self.design['ra'][ev],
                                 dec=self.design['dec'][ev],
                                 delta_ra=self.design['delta_ra'][ev],
                                 delta_dec=self.design['delta_dec'][ev])
        self.design['ra_off'][ev], self.design['dec_off'][ev] = res
        fieldWarn = self.radec_to_xy(ev)

        if np.any(fieldWarn):
            flag = ('Coordio xy coordinates converted '
                    'should be eyed with suspicion.')
            warnings.warn(flag, MugatuDesignWarning)

        self.design_built = True

        return

    def build_design_manual(self):
        """
        Populate the design dictonary for manual design.

        Notes
        -----
        This function creates a manual design whether it is from
        user inputted catalogids, or if it is a FITS file
        if manual_design=True.

        """

        if not self.manual_design:
            flag = 'Building manual design with manual_design=False flag'
            warnings.warn(flag, MugatuWarning)

        # initialize dict for the design
        # for manual design, dont make arrays since length
        # will be defined by user
        self.design['design_pk'] = self.design_pk
        self.design['desmode_label'] = self.desmode_label

        if self.design_file is None:
            # manual design with targets in targetdb
            self.design['catalogID'] = self.catalogids
            self.design['obsWavelength'] = self.obsWavelength
            self.design['priority'] = self.priority
            self.design['carton_pk'] = self.carton_pk
            self.design['category'] = self.category

            if self.ra is None:
                for i in range(len(self.design['catalogID'])):
                    targ_db = Target.get(catalogid=self.design['catalogID'][i])
                    self.design['ra'][i] = targ_db.ra
                    self.design['dec'][i] = targ_db.dec
                    self.design['pmra'][i] = targ_db.pmra
                    self.design['pmdec'][i] = targ_db.pmdec
                    self.design['delta_ra'][i] = targ_db.delta_ra
                    self.design['delta_dec'][i] = targ_db.delta_dec
            else:
                self.design['ra'] = self.ra
                self.design['dec'] = self.dec
                self.design['pmra'] = self.pmra
                self.design['pmdec'] = self.pmdec
                self.design['delta_ra'] = self.delta_ra
                self.design['delta_dec'] = self.delta_dec
                self.design['epoch'] = self.epoch

            self.design['robotID'] = self.robotID
            self.design['holeID'] = np.zeros(len(self.design['robotID']),
                                             dtype='<U10')
            for i in range(len(self.design['robotID'])):
                self.design['holeID'][i] = self.holeID_mapping[self.design['robotID'][i] - 1]
            self.design['magnitudes'] = self.magnitudes
        else:
            # manual design from flat file
            # get header with field info
            head = fits.open(self.design_file)[0].header
            desmode_labels = head['DESMODE'].split(' ')
            # association between catalogid and instrument
            design_inst = fits.open(self.design_file)[1].data
            # catalogid assignment for each fiber
            design = fits.open(self.design_file)[2].data

            # grab obs info
            self.racen = head['RACEN']
            self.deccen = head['DECCEN']
            self.position_angle = head['PA']
            self.observatory = head['obs'].strip().upper()
            self.desmode_label = desmode_labels[self.exp - 1]

            # grab assignment info
            if self.exp == 0:
                roboIDs = design['robotID']
                holeIDs = design['holeID']
            else:
                roboIDs = design['robotID'][:, self.exp - 1]
                holeIDs = design['holeID'][:, self.exp - 1]
            if self.idtype == 'catalogID':
                self.design['catalogID'] = (design_inst['catalogid']
                                            [roboIDs != -1])
            else:
                self.design['catalogID'] = (design_inst['carton_to_target_pk']
                                            [roboIDs != -1])
            self.design['ra'] = design_inst['ra'][roboIDs != -1]
            self.design['dec'] = design_inst['dec'][roboIDs != -1]
            self.design['delta_ra'] = design_inst['delta_ra'][roboIDs != -1]
            self.design['delta_dec'] = design_inst['delta_dec'][roboIDs != -1]
            self.design['pmra'] = design_inst['pmra'][roboIDs != -1]
            self.design['pmdec'] = design_inst['pmdec'][roboIDs != -1]
            self.design['epoch'] = design_inst['epoch'][roboIDs != -1]
            self.design['holeID'] = holeIDs[roboIDs != -1]
            self.design['robotID'] = np.zeros(len(self.design['holeID']),
                                              dtype=int) - 1
            for i in range(len(self.design['holeID'])):
                self.design['robotID'][i] = self.holeID_to_robotID(
                    self.design['holeID'][i],
                    return_ind=False)
            self.design['obsWavelength'] = (design_inst['fiberType']
                                            [roboIDs != -1])
            self.design['priority'] = design_inst['priority'][roboIDs != -1]
            self.design['carton_pk'] = design_inst['carton_pk'][roboIDs != -1]
            self.design['category'] = design_inst['category'][roboIDs != -1]
            self.design['magnitudes'] = design_inst['magnitude'][roboIDs != -1]

        # set nan pm tp zero
        self.design['pmra'][np.isnan(self.design['pmra'])] = 0.
        self.design['pmdec'][np.isnan(self.design['pmdec'])] = 0.
        # make empty x,y arrays
        self.design['x'] = (np.zeros(len(self.design['catalogID']),
                            dtype=float) -
                            9999.99)
        self.design['y'] = (np.zeros(len(self.design['catalogID']),
                            dtype=float) -
                            9999.99)
        self.design['ra_off'] = (np.zeros(len(self.design['catalogID']),
                                 dtype=float) -
                                 9999.99)
        self.design['dec_off'] = (np.zeros(len(self.design['catalogID']),
                                  dtype=float) -
                                  9999.99)

        # here convert ra/dec to x/y based on field/time of observation
        ev = eval("(self.design['ra'] != -9999.99)")
        res = self._offset_radec(ra=self.design['ra'][ev],
                                 dec=self.design['dec'][ev],
                                 delta_ra=self.design['delta_ra'][ev],
                                 delta_dec=self.design['delta_dec'][ev])
        self.design['ra_off'][ev], self.design['dec_off'][ev] = res
        fieldWarn = self.radec_to_xy(ev)

        if np.any(fieldWarn):
            flag = ('Coordio xy coordinates converted '
                    'should be eyed with suspicion.')
            warnings.warn(flag, MugatuDesignWarning)

        self.design_built = True

        return

    def design_to_RobotGrid(self):
        """
        Add assignments to Kaiju RobotGrid.

        Notes
        -----
        Adds targets to the kaiju.robotGrid.RobotGridFilledHex and
        assigns them to fibers based on the design dictonary.

       """

        is_unassigned = False

        for i in range(len(self.design['x'])):
            if self.design['robotID'][i] != -1:
                if self.design['obsWavelength'][i] == 'BOSS':
                    self.rg.addTarget(targetID=self.design['catalogID'][i],
                                      xyzWok=[self.design['x'][i],
                                              self.design['y'][i],
                                              POSITIONER_HEIGHT],
                                      priority=self.design['priority'][i],
                                      fiberType=kaiju.cKaiju.BossFiber)
                else:
                    self.rg.addTarget(targetID=self.design['catalogID'][i],
                                      xyzWok=[self.design['x'][i],
                                              self.design['y'][i],
                                              POSITIONER_HEIGHT],
                                      priority=self.design['priority'][i],
                                      fiberType=kaiju.cKaiju.ApogeeFiber)
        for i in range(len(self.design['x'])):
            if self.design['robotID'][i] > 0:
                try:
                    self.rg.assignRobot2Target(self.design['robotID'][i],
                                               self.design['catalogID'][i])
                except RuntimeError:
                    # this catches the fact that robot cant be
                    # assigned to given fiber
                    self.targets_unassigned.append(self.design['catalogID'][i])
                    is_unassigned = True

        if is_unassigned:
            flag = 'Some targets could not be assigned to fiber'
            warnings.warn(flag, MugatuDesignWarning)

        return

    def decollide_grid(self):
        """
        Check to see if any collisions in grid, and if so,
        decollide grid and record assignments that were
        removed
        """
        # decollide the unassigned robots
        for robotID in self.rg.robotDict:
            if(self.rg.robotDict[robotID].isAssigned() is False):
                self.rg.decollideRobot(robotID)
                # need to still set alpha/beta
                # robot = self.rg.getRobot(robotID)
                # robot.setDestinationAlphaBeta(0, 180)
        if self.rg.getNCollisions() > 0:
            self.design_errors['no_collisions'] = False
            self.rg.decollideGrid()

            # check if de-collision was successful
            if not self.rg.getNCollisions() == 0:
                raise MugatuDesignError(message='Kaiju decollideGrid failed')

            flag = 'Some targets removed from design due to collisions'
            warnings.warn(flag, MugatuDesignWarning)

            # grab all of the targets removed due to collisions
            for i in self.rg.robotDict:
                fiber_idx = np.where(self.design['robotID'] == i)[0]
                if (self.rg.robotDict[i].assignedTargetID == -1 and
                   len(fiber_idx) > 0):
                    targ_remove = self.design['catalogID'][fiber_idx[0]]
                    if targ_remove not in self.targets_unassigned:
                        self.targets_collided.append(targ_remove)
        else:
            self.design_errors['no_collisions'] = True
        return

    def designmode_validate(self, db_query_results_boss=None,
                            db_query_results_apogee=None,
                            desmode_manual=None):
        """
        Check all designmode parameters for a design
        """
        mode = DesignModeCheck(FPSDesign=self,
                               desmode_label=self.desmode_label,
                               db_query_results_boss=db_query_results_boss,
                               db_query_results_apogee=db_query_results_apogee,
                               desmode_manual=desmode_manual)
        mode.design_mode_check_all(verbose=False)
        self.design_errors['min_skies_boss'] = mode.n_skies_min_check['BOSS']
        self.design_errors['min_skies_boss_metric'] = (mode
                                                       .n_skies_min_check
                                                       ['BOSS_metric'])
        if self.design_errors['min_skies_boss'] is False:
            flag = 'Design does not meet minimum BOSS skies for DesignMode'
            warnings.warn(flag, MugatuDesignModeWarning)
        self.design_errors['min_skies_apogee'] = (mode
                                                  .n_skies_min_check['APOGEE'])
        self.design_errors['min_skies_apogee_metric'] = (mode
                                                         .n_skies_min_check
                                                         ['APOGEE_metric'])
        if self.design_errors['min_skies_apogee'] is False:
            flag = 'Design does not meet minimum APOGEE skies for DesignMode'
            warnings.warn(flag, MugatuDesignModeWarning)

        self.design_errors['fov_skies_boss'] = (mode
                                                .min_skies_fovmetric_check
                                                ['BOSS'])
        self.design_errors['fov_skies_boss_metric'] = (mode
                                                       .min_skies_fovmetric_check
                                                       ['BOSS_metric'])
        if self.design_errors['fov_skies_boss'] is False:
            flag = ('Design does not meet FOV criteria '
                    'for BOSS skies for DesignMode')
            warnings.warn(flag, MugatuDesignModeWarning)
        self.design_errors['fov_skies_apogee'] = (mode
                                                  .min_skies_fovmetric_check
                                                  ['APOGEE'])
        self.design_errors['fov_skies_apogee_metric'] = (mode
                                                         .min_skies_fovmetric_check
                                                         ['APOGEE_metric'])
        if self.design_errors['fov_skies_apogee'] is False:
            flag = ('Design does not meet FOV criteria '
                    'for APOGEE skies for DesignMode')
            warnings.warn(flag, MugatuDesignModeWarning)

        self.design_errors['min_stds_boss'] = mode.n_stds_min_check['BOSS']
        self.design_errors['min_stds_boss_metric'] = (mode
                                                      .n_stds_min_check
                                                      ['BOSS_metric'])
        if self.design_errors['min_stds_boss'] is False:
            flag = 'Design does not meet minimum BOSS standards for DesignMode'
            warnings.warn(flag, MugatuDesignModeWarning)
        self.design_errors['min_stds_apogee'] = mode.n_stds_min_check['APOGEE']
        self.design_errors['min_stds_apogee_metric'] = (mode
                                                        .n_stds_min_check
                                                        ['APOGEE_metric'])
        if self.design_errors['min_stds_apogee'] is False:
            flag = ('Design does not meet minimum '
                    'APOGEE standards for DesignMode')
            warnings.warn(flag, MugatuDesignModeWarning)

        self.design_errors['fov_stds_boss'] = (mode
                                               .min_stds_fovmetric_check
                                               ['BOSS'])
        self.design_errors['fov_stds_boss_metric'] = (mode
                                                      .min_stds_fovmetric_check
                                                      ['BOSS_metric'])
        if self.design_errors['fov_stds_boss'] is False:
            flag = ('Design does not meet FOV criteria '
                    'for BOSS standards for DesignMode')
            warnings.warn(flag, MugatuDesignModeWarning)
        self.design_errors['fov_stds_apogee'] = (mode
                                                 .min_stds_fovmetric_check
                                                 ['APOGEE'])
        self.design_errors['fov_stds_apogee_metric'] = (mode
                                                        .min_stds_fovmetric_check
                                                        ['APOGEE_metric'])
        if self.design_errors['fov_stds_apogee'] is False:
            flag = ('Design does not meet FOV criteria '
                    'for APOGEE standards for DesignMode')
            warnings.warn(flag, MugatuDesignModeWarning)

        self.design_errors['stds_mag_boss'] = np.all(
            mode.stds_mags_check['BOSS'][0][(self.design['catalogID'] != -1) &
                                            (self.design['category'] == 'standard_boss')])
        self.design_errors['stds_mag_boss_metric'] = mode.stds_mags_check['BOSS_metric']
        if self.design_errors['stds_mag_boss'] is False:
            flag = ('Design has BOSS standard '
                    'assignments too bright for DesignMode')
            warnings.warn(flag, MugatuDesignModeWarning)
        self.design_errors['stds_mag_apogee'] = np.all(
            mode.stds_mags_check['APOGEE'][0][(self.design['catalogID'] != -1) &
                                              (self.design['category'] == 'standard_apogee')])
        self.design_errors['stds_mag_apogee_metric'] = mode.stds_mags_check['APOGEE_metric']
        if self.design_errors['stds_mag_apogee'] is False:
            flag = ('Design has APOGEE standard '
                    'assignments too bright for DesignMode')
            warnings.warn(flag, MugatuDesignModeWarning)

        self.design_errors['sci_mag_boss'] = np.all(
            mode.bright_limit_targets_check['BOSS'][0][(self.design['catalogID'] != -1) &
                                                       (self.design['category'] == 'science') &
                                                       (self.design['obsWavelength'] == 'BOSS')])
        self.design_errors['sci_mag_boss_metric'] = mode.bright_limit_targets_check['BOSS_metric']
        if self.design_errors['sci_mag_boss'] is False:
            flag = ('Design has BOSS science assignments '
                    ' too bright for DesignMode')
            warnings.warn(flag, MugatuDesignModeWarning)
        self.design_errors['sci_mag_apogee'] = np.all(
            mode.bright_limit_targets_check['APOGEE'][0][(self.design['catalogID'] != -1) &
                                                         (self.design['category'] == 'science') &
                                                         (self.design['obsWavelength'] ==
                                                          'APOGEE')])
        self.design_errors['sci_mag_apogee_metric'] = (mode
                                                       .bright_limit_targets_check
                                                       ['APOGEE_metric'])
        if self.design_errors['sci_mag_apogee'] is False:
            flag = ('Design has APOGEE science assignments '
                    'too bright for DesignMode')
            warnings.warn(flag, MugatuDesignModeWarning)

        self.design_errors['bright_neigh_boss'] = np.all(
            mode.bright_neighbor_check['BOSS'][0][mode.bright_neighbor_check['BOSS'][1] &
                                                  mode.bright_neighbor_check['BOSS'][3]])
        self.design_errors['bright_neigh_boss_metric'] = mode.bright_neighbor_check['BOSS_metric']
        if self.design_errors['bright_neigh_boss'] == False:
            flag = ('Design has BOSS fibers too near '
                    'bright source for DesignMode')
            warnings.warn(flag, MugatuDesignModeWarning)

        self.design_errors['bright_neigh_apogee'] = np.all(
            mode.bright_neighbor_check['APOGEE'][0][mode.bright_neighbor_check['APOGEE'][1] &
                                                    mode.bright_neighbor_check['APOGEE'][3]])
        self.design_errors['bright_neigh_apogee_metric'] = (mode
                                                            .bright_neighbor_check
                                                            ['APOGEE_metric'])
        if self.design_errors['bright_neigh_apogee'] == False:
            flag = ('Design has APOGEE fibers too '
                    'near bright source for DesignMode')
            warnings.warn(flag, MugatuDesignModeWarning)
        return

    def bright_neigh_safety(self, db_query_results_boss=None,
                            db_query_results_apogee=None,
                            desmode_manual=None):
        """
        Perform safety version of bright neighbor check
        on the design
        """
        mode = DesignModeCheck(FPSDesign=self,
                               desmode_label=self.desmode_label,
                               db_query_results_boss=db_query_results_boss,
                               db_query_results_apogee=db_query_results_apogee,
                               desmode_manual=desmode_manual)
        bright_check_boss, hasFiber_boss, _, isassigned_boss = mode.bright_neighbors(
            instrument='BOSS',check_type='safety')
        bright_check_apogee, hasFiber_apogee, _, isassigned_apogee = mode.bright_neighbors(
            instrument='APOGEE', check_type='safety')
        if (len(bright_check_boss[~bright_check_boss & hasFiber_boss & isassigned_boss]) > 0 or
           len(bright_check_apogee[~bright_check_apogee & hasFiber_apogee & isassigned_apogee]) > 0):
            message = 'Bright Neighbor Safety Checked Failed,'
            message += (' %d BOSS and %d APOGEE fibers near bright sources' %
                        (len(bright_check_boss[~bright_check_boss & hasFiber_boss & isassigned_boss]),
                         len(bright_check_apogee[~bright_check_apogee & hasFiber_apogee & isassigned_apogee])))
            raise MugatuDesignError(message=message)
        return

    def RobotGrid_to_valid_design(self):
        """
        Construct valid design from Kaiju Robotgrid
       """

        # initialize dict for the validated design
        # not using None or nan for no assignments
        # using -1 (for int) and -9999.99 (for float) for None assignment
        self.valid_design['design_pk'] = self.design_pk
        self.valid_design['desmode_label'] = self.desmode_label
        self.valid_design['catalogID'] = np.zeros(500, dtype=np.int64) - 1
        self.valid_design['robotID'] = np.zeros(500, dtype=np.int64) - 1
        self.valid_design['holeID'] = np.zeros(500, dtype='<U10')
        self.valid_design['obsWavelength'] = np.zeros(500, dtype='<U6')
        self.valid_design['priority'] = np.zeros(500, dtype=int) - 1
        self.valid_design['carton_pk'] = np.zeros(500, dtype=int) - 1
        self.valid_design['category'] = np.zeros(500, dtype='<U10')
        self.valid_design['ra'] = np.zeros(500, dtype=float) - 9999.99
        self.valid_design['dec'] = np.zeros(500, dtype=float) - 9999.99
        self.valid_design['pmra'] = np.zeros(500, dtype=float) - 9999.99
        self.valid_design['pmdec'] = np.zeros(500, dtype=float) - 9999.99
        self.valid_design['delta_ra'] = np.zeros(500, dtype=float) - 9999.99
        self.valid_design['delta_dec'] = np.zeros(500, dtype=float) - 9999.99
        self.valid_design['ra_off'] = np.zeros(500, dtype=float) - 9999.99
        self.valid_design['dec_off'] = np.zeros(500, dtype=float) - 9999.99
        self.valid_design['epoch'] = np.zeros(500, dtype=float) - 9999.99
        self.valid_design['x'] = np.zeros(500, dtype=float) - 9999.99
        self.valid_design['y'] = np.zeros(500, dtype=float) - 9999.99

        for i, rid in enumerate(self.rg.robotDict):
            self.valid_design['catalogID'][i] = (self.rg.robotDict[rid]
                                                 .assignedTargetID)
            if self.valid_design['catalogID'][i] != -1:
                self.valid_design['robotID'][i] = rid
                ind_hole = np.where(self.robotID_mapping == rid)[0][0]
                self.valid_design['holeID'][i] = self.holeID_mapping[ind_hole]
                # is below necessary? i dont know if decollide ever reassigns
                # or just removes
                cond = eval("self.design['catalogID'] == self.valid_design['catalogID'][i]")
                self.valid_design['obsWavelength'][i] = self.design['obsWavelength'][cond][0]
                self.valid_design['priority'][i] = self.design['priority'][cond][0]
                self.valid_design['carton_pk'][i] = self.design['carton_pk'][cond][0]
                self.valid_design['category'][i] = self.design['category'][cond][0]
                self.valid_design['ra'][i] = self.design['ra'][cond][0]
                self.valid_design['dec'][i] = self.design['dec'][cond][0]
                self.valid_design['pmra'][i] = self.design['pmra'][cond][0]
                self.valid_design['pmdec'][i] = self.design['pmdec'][cond][0]
                self.valid_design['delta_ra'][i] = self.design['delta_ra'][cond][0]
                self.valid_design['delta_dec'][i] = self.design['delta_dec'][cond][0]
                self.valid_design['ra_off'][i] = self.design['ra_off'][cond][0]
                self.valid_design['dec_off'][i] = self.design['dec_off'][cond][0]
                self.valid_design['epoch'][i] = self.design['epoch'][cond][0]
            self.valid_design['x'][i] = self.rg.robotDict[rid].xPos
            self.valid_design['y'][i] = self.rg.robotDict[rid].yPos

        return

    def validate_design(self, designmode=True, safety=True,
                        db_query_results_boss=None,
                        db_query_results_apogee=None,
                        desmode_manual=None):
        """
        Validate design for deadlocks and collisions using Kaiju.

        Parameters
        ----------
        designmode: bool
            Check designmodes for the design.

        safety: bool
            Check for bright neighbors using carton of brightest Gaia/2MASS
            stars in targetdb.

        db_query_results_boss: dict
            Database query results for BOSS bright neighbor check.
            Each index of dict is a tuple of (ras, decs, mags, catalogids)
            with one index for designmode and the other safety.

        db_query_results_apogee: dict
            Database query results for APOGEE bright neighbor check.
            Each index of dict is a tuple of (ras, decs, mags, catalogids)
            with one index for designmode and the other safety.
        """
        # make dict to store design errors that may come up
        self.design_errors = {}
        # build the design with all needed parameters if it has not been
        # built already (save time by doing this check)
        if not self.design_built:
            if self.manual_design:
                self.build_design_manual()
            else:
                self.build_design_db()

        # construct the Kaiju robotGrid
        self.design_to_RobotGrid()
        if len(self.targets_unassigned) > 0:
            self.design_errors['all_targets_assigned'] = False
        else:
            self.design_errors['all_targets_assigned'] = True

        # validate the design

        # de-collide the grid if collisions exist
        # and check for targets removed
        self.decollide_grid()

        # generate paths
        # self.rg.pathGen()
        # if self.rg.didFail:
        #     raise MugatuError(message='Kaiju pathGen failed')

        if designmode:
            self.designmode_validate(db_query_results_boss=db_query_results_boss,
                                     db_query_results_apogee=db_query_results_apogee,
                                     desmode_manual=desmode_manual)

        # do the safety check for design
        if safety:
            self.bright_neigh_safety(db_query_results_boss=db_query_results_boss,
                                     db_query_results_apogee=db_query_results_apogee,
                                     desmode_manual=desmode_manual)

        # I imagine that the above step would manipulate the robogrid based on
        # collisions and deadlocks, so the below would take these Kaiju results
        # and put them back as a design object which will be the output of the
        # function
        self.RobotGrid_to_valid_design()

        # another note to add to this above design dictonary, is how will this
        # include paths? This seems important for manual designs and to send
        # to Jaeger

        return

    def design_to_targetdb(self, cadence, fieldid, exposure):
        """
        Write a validated esign to targetdb

        Parameters
        ----------

        cadence: str
            Cadence label for the field.

        fieldid: int
            fieldid for the field.

        exposure: int
            The exposure of this set of designs, 0th indexed.

        Notes
        -----
        Version above should allow for manual designs to be added under
        version = 'manual' to seperate them from robostrategy ingested designs
        """
        if not self.manual_design:
            raise MugatuError(message='Can only write manual designs to targetdb')

        if len(self.valid_design) == 0:
            raise MugatuError(message='Need to validate design first')

        # create the manual field for the design
        make_design_field_targetdb(cadence=cadence,
                                   fieldid=fieldid,
                                   plan='manual',
                                   racen=self.racen,
                                   deccen=self.deccen,
                                   position_angle=self.position_angle,
                                   observatory=self.observatory,
                                   slots_exposures=None)

        # create dictonary for unique carton pks for idtype is catalogID
        if self.idtype == 'catalogID':
            cart_pks = {}
            targetdb_ver = {}
            for pk in np.unique(self.valid_design['carton_pk'][self.valid_design['catalogID'] != -1]):
                cart_pks[pk] = pk
                targetdb_ver[pk] = Carton.get(pk).version_pk
        else:
            cart_pks = None
            targetdb_ver = None

        # add the design to targetdb
        make_design_assignments_targetdb(
            plan='manual',
            fieldid=fieldid,
            exposure=exposure,
            desmode_label=self.desmode_label,
            catalogID=self.valid_design['catalogID'][self.valid_design['catalogID'] != -1],
            robotID=self.valid_design['robotID'][self.valid_design['catalogID'] != -1],
            holeID=self.valid_design['holeID'][self.valid_design['catalogID'] != -1],
            obsWavelength=self.valid_design['obsWavelength'][self.valid_design['catalogID'] != -1],
            carton=self.valid_design['carton_pk'][self.valid_design['catalogID'] != -1],
            observatory=self.observatory,
            targetdb_ver=targetdb_ver,
            instr_pks=None,
            cart_pks=cart_pks,
            fiber_pks=None,
            idtype=self.idtype)
        return

    def design_to_opsdb(self, design):
        """
        Write a validated design to opsdb

        Notes
        -----
        I dont know what needs to be included in opsdb yet, need to look
        at the schema
        """
        return



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'python'}

packages = \
['mugatu']

package_data = \
{'': ['*'],
 'mugatu': ['etc/*',
            'templates/dmode_check_w_dists.html',
            'templates/dmode_check_w_dists.html',
            'templates/dmode_check_w_dists.html',
            'templates/dmode_check_w_dists.html',
            'templates/dmode_check_w_dists.html',
            'templates/dmode_check_w_dists_sky.html',
            'templates/dmode_check_w_dists_sky.html',
            'templates/dmode_check_w_dists_sky.html',
            'templates/dmode_check_w_dists_sky.html',
            'templates/dmode_check_w_dists_sky.html',
            'templates/main_validation_page.html',
            'templates/main_validation_page.html',
            'templates/main_validation_page.html',
            'templates/main_validation_page.html',
            'templates/main_validation_page.html',
            'templates/valid_check.html',
            'templates/valid_check.html',
            'templates/valid_check.html',
            'templates/valid_check.html',
            'templates/valid_check.html',
            'templates/valid_check_w_dists.html',
            'templates/valid_check_w_dists.html',
            'templates/valid_check_w_dists.html',
            'templates/valid_check_w_dists.html',
            'templates/valid_check_w_dists.html']}

install_requires = \
['DateTime>=4.3,<5.0',
 'fitsio>=1.0.5,<2.0.0',
 'jupyter>=1.0.0,<2.0.0',
 'numpy>=1.19.5,<2.0.0',
 'ortools==9.1.9490',
 'scipy>=1.6.0,<2.0.0',
 'sdss-access==1.1.1',
 'sdss-coordio==1.2.1',
 'sdss-kaiju==1.2.2',
 'sdssdb>=0.5.0,<0.6.0']

setup_kwargs = {
    'name': 'sdss-mugatu',
    'version': '1.1.2',
    'description': 'Package to read, write and validate FPS designs',
    'long_description': '# mugatu\n\n![Versions](https://img.shields.io/badge/python->3.7-blue)\n[![Documentation Status](https://readthedocs.org/projects/sdss-mugatu/badge/?version=latest)](https://sdss-mugatu.readthedocs.io/en/latest/?badge=latest)\n[![Travis (.org)](https://img.shields.io/travis/sdss/mugatu)](https://travis-ci.org/sdss/mugatu)\n[![codecov](https://codecov.io/gh/sdss/mugatu/branch/main/graph/badge.svg)](https://codecov.io/gh/sdss/mugatu)\n\nPackage to read, write and validate FPS designs\n',
    'author': 'Ilija Medan',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/sdss-mugatu/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)
